class QasePytestOptions:
    
    @staticmethod
    def addoptions(parser, group, added=[]):

        QasePytestOptions.add_option_ini(
            parser,
            group,
            "--qase-mode",
            dest="qase_mode",
            default="off",
            help="Define Qase reporter mode: `off`, `report` or `testops`"
        )

        QasePytestOptions.add_option_ini(
            parser,
            group,
            "--qase-environment",
            dest="qase_environment",
            help="Define environment slug or ID from TestOps"
        )

        QasePytestOptions.add_option_ini(
            parser,
            group,
            "--qase-debug",
            dest="qase_debug",
            type="bool",
            default=False,
            help="Enable debug mode"
        )

        QasePytestOptions.add_option_ini(
            parser,
            group,
            "--qase-execution-plan-path",
            dest="qase_execution_plan_path",
            default="qase_execution_plan.json",
            help="Path to file with execution plan"
        )

        QasePytestOptions.add_option_ini(
            parser,
            group,
            "--qase-testops-project",
            dest="qase_testops_project",
            help="Project code in Qase TestOps"
        )

        QasePytestOptions.add_option_ini(
            parser,
            group,
            "--qase-testops-api-token",
            dest="qase_testops_api_token",
            help="API token for Qase TestOps"
        )

        QasePytestOptions.add_option_ini(
            parser,
            group,
            "--qase-testops-api-host",
            dest="qase_testops_api_host",
            default="qase.io",
            help="API host for Qase TestOps"
        )

        QasePytestOptions.add_option_ini(
            parser,
            group,
            "--qase-testops-plan-id",
            dest="qase_testops_plan_id",
            help="Test Plan ID in Qase TestOps"
        )

        QasePytestOptions.add_option_ini(
            parser,
            group,
            "--qase-testops-run-id",
            dest="qase_testops_run_id",
            help="Test Run ID in Qase TestOps"
        )

        QasePytestOptions.add_option_ini(
            parser,
            group,
            "--qase-testops-run-title",
            "qase_testops_run_title",
            help="Define title for autocreated Test Run. If not set, will be used autogenerated title"
        )

        QasePytestOptions.add_option_ini(
            parser,
            group,
            "--qase-testops-run-description",
            dest="qase_testops_run_description",
            default="",
            help="Define description for autocreated Test Run"
        )

        QasePytestOptions.add_option_ini(
            parser,
            group,
            "--qase-testops-run-complete",
            dest="qase_testops_run_description",
            type="bool",
            default=False,
            help="Complete run after tests execution"
        )

        QasePytestOptions.add_option_ini(
            parser,
            group,
            "--qase-testops-bulk",
            dest="qase_testops_bulk",
            type="bool",
            default=True,
            help="Send results in bulk after tests execution or one by one after each test"
        )

        QasePytestOptions.add_option_ini(
            parser,
            group,
            "--qase-testops-batch",
            dest="qase_testops_batch",
            default=200,
            help="Define batch size for bulk sending. Default: 200"
        )

        QasePytestOptions.add_option_ini(
            parser,
            group,
            "--qase-testops-defect",
            dest="qase_testops_defect",
            type="bool",
            default=False,
            help="Create defect if test failed"
        )

        QasePytestOptions.add_option_ini(
            parser,
            group,
            "--qase-report-driver",
            dest="qase_report_driver",
            default="local",
            help="Define report driver: `local`. More options coming soon."
        )

        QasePytestOptions.add_option_ini(
            parser,
            group,
            "--qase-report-connection-local-path",
            dest="qase_report_connection_local_path",
            default="./build/qase-report",
            help="Define local path for report directory"
        )

        QasePytestOptions.add_option_ini(
            parser,
            group,
            "--qase-report-connection-local-format",
            dest="qase_report_connection_local_format",
            default="json",
            help="Define local format for report directory: `json` or `jsonp`"
        )

        QasePytestOptions.add_option_ini(
            parser,
            group,
            "--qase-framework-pytest-capture-logs",
            dest="qase_framework_pytest_capture_logs",
            type="bool",
            default=False,
            help="Capture logs on fail for each test"
        )

        QasePytestOptions.add_option_ini(
            parser,
            group,
            "--qase-framework-pytest-capture-http",
            dest="qase_framework_pytest_capture_http",
            type="bool",
            default=False,
            help="Capture http requests for each test and save them as steps"
        )
    
    @staticmethod
    def add_option_ini(parser, group, option, dest, default=None, type=None, **kwargs):
        # We are going to add options that were not added before through the manager
        try:
            parser.addini(
                dest,
                default=default,
                type=type,
                help="default value for " + option,
            )
            group.addoption(option, dest=dest, default=default, action="store", **kwargs)
        except ValueError:
            pass


from qaseio.pytest.plugin import QasePytestPlugin, QasePytestPluginSingleton

from qaseio.commons import QaseTestOps, QaseReport, ConfigManager
from qaseio.pytest.options import QasePytestOptions

import os
import json

def pytest_addoption(parser):
    group = parser.getgroup("qase")

    config_manager = ConfigManager()
    config_manager.load_config()

    for key in config_manager._get_keys(config_manager.config):
        default_value = config_manager.get(key)
        key_arg = key.replace("_", "-")
        group.addoption(f"--qase-{key_arg}", action="store", default=default_value)

    QasePytestOptions.addoptions(parser, group)

def pytest_configure(config):
    if not hasattr(config, "workerinput"):
        QasePytestPlugin.drop_run_id()
    config.addinivalue_line("markers", "qase_id: mark test to be associate with Qase TestOp test case")
    config.addinivalue_line("markers", "qase_title: mark test with title")
    config.addinivalue_line("markers", "qase_ignore: skip test from Qase TestOps \ Report")
    config.addinivalue_line("markers", "qase_muted: mark test as muted so it will not affect test run status")
    config.addinivalue_line("markers", "qase_author: mark test with author")
    config.addinivalue_line("markers", "qase_fields: mark test with meta data")
    config.addinivalue_line("markers", "qase_suite: mark test with suite")

    # Legacy markers
    config.addinivalue_line("markers", "qase_description: mark test with description")
    config.addinivalue_line("markers", "qase_preconditions: mark test with preconditions")
    config.addinivalue_line("markers", "qase_postconditions: mark test with postconditions")
    config.addinivalue_line("markers", "qase_layer: mark test with layer")
    config.addinivalue_line("markers", "qase_severity: mark test with severity")
    config.addinivalue_line("markers", "qase_priority: mark test with priority")
    config.addinivalue_line("markers", "qase_tags: mark test with tags")

    mode = config.getoption("qase_mode", None)
    execution_plan = None

    # Load execution plan from file
    path = config.getoption("qase_execution_plan_path", "qase_execution_plan.json")
    if not config.getoption("qase_execution_plan_path", None) and path and os.path.isfile(path):
        with open('execution_plan.json') as f:
            execution_plan = json.load(f)

    # Load execution plan from command line or env variable
    if config.getoption("qase_execution_plan", None):
        execution_plan = [int(n) for n in str(config.getoption("qase_execution_plan").split(","))]

    if mode:
        defaultReporter = QaseReport(
                report_path=config.getoption("qase_report_connection_local_path", "./build/qase-report"),
                format=config.getoption("qase_report_connection_local_format", "json"),
                environment=config.getoption("qase_environment", "local"),
            )
        if (mode == 'testops'):
            if validate_testops_options(config):
                if (config.getoption("qase_testops_plan_id", None) is not None):
                    from qaseio.commons import TestOpsPlanLoader

                    # Load test plan data from Qase TestOps
                    loader = TestOpsPlanLoader(
                        api_token=config.getoption("qase_testops_api_token"),
                        host=config.getoption("qase_testops_api_host", "qase.io"),
                    )
                    execution_plan = loader.load(config.getoption("qase_testops_project"),
                                                 int(config.getoption("qase_testops_plan_id")))

                reporter = QaseTestOps(
                    api_token=config.getoption("qase_testops_api_token"),
                    project_code=config.getoption("qase_testops_project"),
                    run_id=config.getoption("qase_testops_run_id", None),
                    plan_id=config.getoption("qase_testops_plan_id", None),
                    complete_run=config.getoption("qase_testops_run_complete", False),
                    bulk=config.getoption("qase_testops_bulk", True),
                    run_title=config.getoption("qase_testops_run_title", None),
                    host=config.getoption("qase_testops_api_host", "qase.io"),
                    chunk_size=config.getoption("qase_testops_chunk", 200),
                    environment=config.getoption("qase_environment", None),
                    defect=config.getoption("qase_testops_defect", False),
                )
                fallback = defaultReporter
            else:
                print('⚠️  Switching to local report mode')
                reporter = defaultReporter
                fallback = None
        else:
            reporter = defaultReporter
            fallback = None

        QasePytestPluginSingleton.init(
            reporter=reporter,
            fallback=fallback,
            xdist_enabled=is_xdist_enabled(config),
            capture_logs=config.getoption("qase_framework_pytest_capture_logs", False),
            intercept_requests=config.getoption("qase_framework_pytest_capture_http", False),
            execution_plan=execution_plan,
        )
        config.qaseio = QasePytestPluginSingleton.get_instance()
        config.pluginmanager.register(
            config.qaseio,
            name="qase-pytest",
        ) 

def validate_testops_options(config) -> bool:
    if not config.getoption("qase_testops_api_token", None):
        print("[QASE] ⚠️  Qase TestOps API token is required")
        return False
    
    if not config.getoption("qase_testops_project", None):
        print("[QASE] ⚠️  Qase TestOps project code is required")
        return False

    return True

def is_xdist_enabled(config):
    if (config.pluginmanager.getplugin("xdist") is not None and os.getenv('PYTEST_XDIST_WORKER_COUNT') is not None):
        return True
    return False

def pytest_unconfigure(config):
    qaseio = getattr(config, "src", None)
    if qaseio:
        del config.qaseio
        config.pluginmanager.unregister(qaseio)



import pathlib
from typing import Tuple, Union
import pytest
import mimetypes
import re

from qaseio.commons.models.result import Result, Field
from qaseio.commons.models.attachment import Attachment
from qaseio.commons.utils import QaseUtils
from qaseio.commons.models.runtime import Runtime

from filelock import FileLock

PYTEST_TO_QASE_STATUS = {
    "PASSED": 'passed',
    "FAILED": 'failed',
    "SKIPPED": 'skipped',
    "BLOCKED": 'blocked',
    "BROKEN": 'invalid'
}

try:
    from xdist import is_xdist_controller
except ImportError:
    def is_xdist_controller(*args, **kwargs):
        return True
    
try:
    import pytest
except ImportError:
    raise ImportError("pytest is not installed")

class PluginNotInitializedException(Exception):
    pass

class QasePytestPlugin:
    run = None
    meta_run_file = pathlib.Path("src.run")

    def __init__(
            self,
            reporter,
            fallback,
            xdist_enabled = False,
            capture_logs = False,
            execution_plan = None,
            intercept_requests = False,
    ):
        self.runtime = Runtime()
        self.reporter = reporter
        self.run_id = None
        self.xdist_enabled = xdist_enabled
        self.fallback = fallback
        self.capture_logs = capture_logs
        self.debug = True
        self.execution_plan = execution_plan
        self.intercept_requests = intercept_requests
        self.interceptor = None
        if self.intercept_requests:
            # Lazy import
            from qaseio.commons.interceptor import InterceptorSingleton
            InterceptorSingleton.init(runtime=self.runtime)
            self.interceptor = InterceptorSingleton.get_instance()

    def start_step(self, step):
        self.runtime.add_step(step)

    def finish_step(self, id, exception=None):
        status = PYTEST_TO_QASE_STATUS['PASSED']
        if exception:
            status = PYTEST_TO_QASE_STATUS['FAILED']

        self.runtime.finish_step(id, status=status)

    @staticmethod
    def drop_run_id():
        if QasePytestPlugin.meta_run_file.exists():
            QasePytestPlugin.meta_run_file.unlink()

    def pytest_collection_modifyitems(self, session, config, items):
        # Filter test cases based on ids
        if (self.execution_plan):
            items[:] = [item for item in items if item.get_closest_marker('qase_id') and item.get_closest_marker('qase_id').kwargs.get("id") in self.execution_plan]

    def pytest_sessionstart(self, session):
        if (not self.xdist_enabled) or (self.xdist_enabled and is_xdist_controller(session)):
            self.run_id = self.reporter.start_run()
            with FileLock("qaseio.lock"):
                if self.run_id:
                    with open(self.meta_run_file, "w") as lock_file:
                        lock_file.write(str(self.run_id))
        else:
            self.load_run_from_lock()
        self.has_started = True

    def pytest_sessionfinish(self, session, exitstatus):
        main = False
        if (not self.xdist_enabled) or (self.xdist_enabled and is_xdist_controller(session)):
            main = True
            QasePytestPlugin.drop_run_id()
        self.reporter.complete_run(main)
        self.has_started = False

    @pytest.hookimpl(hookwrapper=True)
    def pytest_runtest_protocol(self, item):
        self.ignore = True if item.get_closest_marker("qase_ignore") else False

        if not self.ignore:
            # Capture HTTP requests inside test
            if self.interceptor:
                self.interceptor.enable()
            self.start_pytest_item(item)
            yield
            if self.interceptor:
                self.interceptor.disable()
            self.finish_pytest_item(item)
        else:
            yield

    @pytest.hookimpl(hookwrapper=True)
    def pytest_runtest_makereport(self, item, call):
        if not self.ignore:
            report = (yield).get_result()
            def set_result(res):
                self.runtime.result.execution.status = res

            if report.longrepr:
                self.runtime.result.execution.stacktrace = report.longreprtext

            if report.failed:
                if call.excinfo.typename != "AssertionError":
                    set_result(PYTEST_TO_QASE_STATUS['BROKEN'])
                else:
                    set_result(PYTEST_TO_QASE_STATUS['FAILED'])
                self.runtime.result.add_message(call.excinfo.exconly())
            elif report.skipped:
                if self.runtime.result.execution.status in (
                        None,
                        PYTEST_TO_QASE_STATUS['PASSED'],
                ):
                    set_result(PYTEST_TO_QASE_STATUS['SKIPPED'])
            else:
                if self.runtime.result.execution.status is None:
                    set_result(PYTEST_TO_QASE_STATUS['PASSED'])

            if self.capture_logs and report.when == "call":
                if report.caplog:
                    self.add_attachments((report.caplog, "text/plain", "log.txt"))
                if report.capstdout:
                    self.add_attachments((report.capstdout, "text/plain", "stdout.txt"))
                if report.capstderr:
                    self.add_attachments((report.capstderr, "text/plain", "stderr.txt"))
        else:
            yield

    def start_pytest_item(self, item):
        self.runtime.result = Result(
            title = self._get_title(item), 
            signature = self._get_signature(item),
            )
        self._set_fields(item)
        self._set_tags(item)
        self._set_author(item)
        self._set_muted(item)
        self._set_testops_id(item)
        self._set_params(item)

    def finish_pytest_item(self, item):
        self.runtime.result.execution.complete()
        self.runtime.result.add_steps([step for key, step in self.runtime.steps.items()])
        self.reporter.add_result(self.runtime.result)

        self.runtime = Runtime()

    def add_attachments(
            self, *files: Union[str, Tuple[str, str], Tuple[bytes, str, str]]
    ):
        for file in files:
            filename = None
            content = None
            file_path = None

            if isinstance(file, tuple):
                if len(file) == 2:
                    file_path, mime = file
                    filename = QaseUtils.get_filename(file_path)
                else:
                    content, mime, filename = file
            elif isinstance(file, str):
                file_path = file
                mime = mimetypes.guess_type(file)[0]
                filename = QaseUtils.get_filename(file_path)

            attachment = Attachment(file_name=filename, content=content, mime_type=mime, file_path=file_path)
            self.runtime.add_attachment(attachment)

    def load_run_from_lock(self):
        if QasePytestPlugin.meta_run_file.exists():
            with open(QasePytestPlugin.meta_run_file, "r") as lock_file:
                try:
                    self.run_id = str(lock_file.read())
                    self.reporter.set_run_id(self.run_id)
                except ValueError:
                    pass
    
    def _get_title(self, item):
        title = None
        try:
            title = item.get_closest_marker("qase_title").kwargs.get("title")
        except:
            pass

        if not title:
            title = item.originalname

        return str(title)
    
    def _get_signature(self, item) -> str:
        return re.sub(r'\[.*?\]', '', item.nodeid)
    
    def _set_relations(self, item) -> None:
        # TODO: Add support for relations
        pass
    
    def _set_fields(self, item) -> None:
        # Legacy fields support
        for name in ["description", "preconditions", "postconditions", "layer", "severity", "priority"]:
            try:
                self.runtime.result.add_field(Field(name, item.get_closest_marker("qase_" + name).kwargs.get(name)))
            except:
                pass

        try:
            fields = item.get_closest_marker("qase_fields").kwargs.get("fields")
            for name, field in fields:
                self.runtime.result.add_field(Field(name, field))
        except:
            pass

    def _set_tags(self, item) -> None:
        try:
            tags = item.get_closest_marker("qase_tags").kwargs.get("tags")
            for tag in tags:
                self.runtime.result.add_tag(tag)
        except:
            pass

    def _set_author(self, item) -> None:
        try:
            self.runtime.result.author = str(item.get_closest_marker("qase_author").kwargs.get("author"))
        except:
            pass

    def _set_muted(self, item) -> None:
        try:
            self.runtime.result.muted = True if item.get_closest_marker("qase_muted").kwargs.get("muted") else False
        except:
            pass

    def _set_testops_id(self, item) -> None:
        try:
            self.runtime.result.testops_id = int(item.get_closest_marker("qase_id").kwargs.get("id"))
        except:
            pass

    def _set_params(self, item) -> None:
        if hasattr(item, 'callspec'):
            for key, val in item.callspec.params.items():
                self.runtime.result.add_param(key, str(val))

class QasePytestPluginSingleton:
    _instance = None

    @staticmethod
    def init(**kwargs):
        if QasePytestPluginSingleton._instance is None:
            QasePytestPluginSingleton._instance = QasePytestPlugin(**kwargs)

    @staticmethod
    def get_instance() -> QasePytestPlugin:
        """ Static access method"""
        if QasePytestPluginSingleton._instance is None:
            raise PluginNotInitializedException("Init plugin first")
        return QasePytestPluginSingleton._instance

    def __init__(self):
        """ Virtually private constructor"""
        raise Exception("Use get_instance()")



# -*- coding: utf-8 -*-
import functools
from contextlib import ContextDecorator as PyContextDecorator
from contextlib import _GeneratorContextManager as GeneratorContextManager
import uuid
from pkg_resources import DistributionNotFound, get_distribution
from typing import Tuple, Union

import pytest

from qaseio.pytest.plugin import PluginNotInitializedException, QasePytestPluginSingleton
from qaseio.commons.models.step import Step, StepTextData

try:
    # Change here if project is renamed and does not equal the package name
    dist_name = "qase-pytest"
    __version__ = get_distribution(dist_name).version
except DistributionNotFound:
    __version__ = "unknown"
finally:
    del get_distribution, DistributionNotFound


class ContextManager(GeneratorContextManager, PyContextDecorator):
    """Pass in a generator to the initializer and the resultant object
    is both a decorator closure and context manager
    """

    def __init__(self, func, args=(), kwargs=None):
        if kwargs is None:
            kwargs = {}

        super().__init__(func, args, kwargs)


def contextdecorator(func):
    """Similar to contextlib.contextmanager except the decorated generator
    can be used as a decorator with optional arguments.
    """

    @functools.wraps(func)
    def helper(*args, **kwargs):
        is_decorating = len(args) == 1 and callable(args[0])

        if is_decorating:
            new_func = args[0]

            @functools.wraps(new_func)
            def new_helper(*args, **kwargs):
                instance = ContextManager(func)
                return instance(new_func)(*args, **kwargs)

            return new_helper
        return ContextManager(func, args, kwargs)

    return helper


class qase:
    """Class with decorators for pytest"""

    @staticmethod
    def id(id):
        """
        Define a persisent connection to Qase TestOps test case.

        >>> @qase.id(1)
        >>> def test_example():
        >>>     pass

        :param id: int id of test case
        :return: pytest.mark instance
        """
        return pytest.mark.qase_id(id=id)

    @staticmethod
    def title(title):
        """
        >>> @qase.title("Sign up")
        >>> def test_example():
        >>>     pass

        :param title: a string with test name
        :return: pytest.mark instance
        """
        return pytest.mark.qase_title(title=title)
    
    @staticmethod
    def fields(*fields: Tuple[str, str]):
        """
        >>> @qase.fields(
                ("priority", "high"),
                ("layer", "API")
                ("custom_field", "custom_value")
            )
        >>> def test_example():
        >>>     pass

        :param name: a string with field name
        :param value: a string with field value
        :return: pytest.mark instance
        """
        return pytest.mark.qase_fields(fields=fields)
    
    @staticmethod
    def suite(title: str, description: str = None):
        """
        >>> @qase.suite("Sign up")
        >>> def test_example():
        >>>     pass

        :param title: a string with suite name. You can use dot notation to create nested suites.
        :param description: a string with suite description
        :return: pytest.mark instance
        """
        return pytest.mark.qase_suite(title=title, description=description)
    
    @staticmethod
    def author(author):
        """
        >>> @qase.author("John Doe")
        >>> def test_example():
        >>>     pass

        :param author: a string with test author
        :return: pytest.mark instance
        """
        return pytest.mark.qase_author(author=author)

    @staticmethod
    def description(description):
        """
        >>> @qase.description("Sign up user using login and password")
        >>> def test_example():
        >>>     pass

        :param description: a string with test full description. Markdown is supported.
        :return: pytest.mark instance
        """
        return pytest.mark.qase_description(description=description)

    @staticmethod
    def preconditions(preconditions):
        """
        >>> @qase.preconditions("Sign up user using login and password")
        >>> def test_example():
        >>>     pass

        :param preconditions: a string with test preconditions. Markdown is supported.
        :return: pytest.mark instance
        """
        return pytest.mark.qase_preconditions(preconditions=preconditions)
    
    @staticmethod
    def postconditions(postconditions):
        """
        >>> @qase.postconditions("Sign up user using login and password")
        >>> def test_example():
        >>>     pass

        :param postconditions: a string with test postconditions. Markdown is supported.
        :return: pytest.mark instance
        """
        return pytest.mark.qase_postconditions(postconditions=postconditions)

    @staticmethod
    def severity(severity):
        """
        >>> @qase.severity("critical")
        >>> def test_example():
        >>>     pass

        :param severity: a string with test severity. Default values: blocker, critical, major, normal, minor, trivial.
        :return: pytest.mark instance
        """
        return pytest.mark.qase_severity(severity=severity)

    @staticmethod
    def priority(priority):
        """
        >>> @qase.priority("Critical")
        >>> def test_example():
        >>>     pass

        :param priority: a string with test priority. Default values: low, medium, high.
        :return: pytest.mark instance
        """
        return pytest.mark.qase_priority(priority=priority)

    @staticmethod
    def layer(layer):
        """
        >>> @qase.layer("E2E")
        >>> def test_example():
        >>>     pass

        :param layer: a string with test layer. Default values: E2E, API, Unit.
        :return: pytest.mark instance
        """
        return pytest.mark.qase_layer(layer=layer)
    
    @staticmethod
    def tags(*tags):
        """
        >>> @qase.tags("tag1", "tag2")
        >>> def test_example():
        >>>     pass

        :param tags: a list of strings with test tags.
        :return: pytest.mark instance
        """
        return pytest.mark.qase_tags(tags=tags)
    
    @staticmethod
    def ignore():
        """
        >>> @qase.ignore()
        >>> def test_example():
        >>>     pass

        :return: pytest.mark instance
        """
        return pytest.mark.qase_ignore()
    
    @staticmethod
    def muted():
        """
        >>> @qase.muted()
        >>> def test_example():
        >>>     pass

        :return: pytest.mark instance
        """
        return pytest.mark.qase_muted()
    
    @staticmethod
    def attach(*files: Union[str, Tuple[str, str], Tuple[bytes, str, str]]):
        """
        Attach files to test results

        `files` could be:
            - str - only `filepath`
            - str, str - `filepath` and `mime-type` for it
            - bytes, str, str - `source` data, `mime-type` and `filename`

        >>> from src.client.models import MimeTypes
        ... qase.attach(
        ...     (driver.get_screenshot_as_png(), MimeTypes.PNG, "page.png")
        ... )
        """
        try:
            plugin = QasePytestPluginSingleton.get_instance()
            plugin.add_attachments(*files)
        except Exception:
            print("Failed to add attachments")
            pass

    @staticmethod
    @contextdecorator
    def step(title, expected=None):
        """
        Step context/decorator

        Usage:

        >>> @qase.step("First step")
        ... def first_step():
        ...     print("smthng")

        >>> with qase.step("Second step"):
        ...     print("smthng")
        """
        plugin = None
        id = str(uuid.uuid4())
        try:
            plugin = QasePytestPluginSingleton.get_instance()
            step = Step(
                step_type = 'text', 
                id = id,
                data = StepTextData(
                    action = title,
                    expected_result = expected if (expected) else None,
                )
            )
            plugin.start_step(step)
            yield
            plugin.finish_step(id=step.id)
        except PluginNotInitializedException:
            yield
        except AttributeError:
            yield
        except Exception as e:
            plugin.finish_step(id=id, exception=e)
            raise e


from qaseio.pytest import qase


def capital_case(x):
    return x.capitalize()

@qase.title('Test should pass')
def test_capital_case():
    assert capital_case('semaphor') == 'Semaphor'


@qase.title("Test should fail")
def test_capital_case2():
    assert capital_case('semaphor') == 'semaphor'